;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
;  seh0.inc
;
;  Structured Exception Handling
;
;  Written by Four-F (four-f@mail.ru)
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

SEH STRUCT 
	SafeOffset		dd	?	; The offset where it's safe to continue execution 
	PrevEsp			dd	?	; the old value in esp 
	PrevEbp			dd	?	; The old value in ebp 
SEH ENDS

SEH_SafePlaceCounter	= 0

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

_try MACRO SafePlace, xHandler:=<DefaultExceptionHandler>
	local sp_lbl

	PUSHCONTEXT ASSUMES    
	assume fs:nothing
	
	IFNDEF seh
		.data
		seh	SEH <>
		.code
	ENDIF
   
	push offset xHandler
	push fs:[0]						; address of next ERR structure
	mov fs:[0], esp					; give FS:[0] the ERR address just made

	IFB <SafePlace>
		sp_lbl TEXTEQU @CatStr(<SEH_SafePlace>, %(SEH_SafePlaceCounter))
		% ECHO sp_lbl
		mov seh.SafeOffset, offset sp_lbl
	ELSE
		mov seh.SafeOffset, offset SafePlace	
	ENDIF
	mov seh.PrevEbp, ebp
	mov seh.PrevEsp, esp

	POPCONTEXT ASSUMES

ENDM

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

_finally MACRO
	local sp_lbl

	sp_lbl TEXTEQU @CatStr(<SEH_SafePlace>, %(SEH_SafePlaceCounter))
	SEH_SafePlaceCounter = SEH_SafePlaceCounter + 1
	sp_lbl:

	PUSHCONTEXT ASSUMES    
	assume fs:nothing
	pop fs:[0]						; restore next ERR structure to FS:[0]
	add esp, sizeof DWORD			; throw away rest of ERR structure
	POPCONTEXT ASSUMES
ENDM

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

.code

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

DefaultExceptionHandler proc uses esi edi pExcept:DWORD, pFrame:DWORD, pContext:DWORD, pDispatch:DWORD

	mov eax, pExcept
	invoke DbgPrint, $CTA0("Exception: %08X at address: %08X"), \
						[eax][EXCEPTION_RECORD.ExceptionCode], \
						[eax][EXCEPTION_RECORD.ExceptionAddress]

	IFNDEF seh
		.data
		seh	SEH <>
		.code
	ENDIF

	lea eax, seh
    push (SEH PTR [eax]).SafeOffset
    push (SEH PTR [eax]).PrevEsp
    push (SEH PTR [eax]).PrevEbp

	mov eax, pContext
    pop (CONTEXT PTR [eax]).regEbp
    pop (CONTEXT PTR [eax]).regEsp
    pop (CONTEXT PTR [eax]).regEip

	; reload context & continue execution
    xor eax, eax						;mov eax, ExceptionContinueExecution

    ret 

DefaultExceptionHandler endp

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
